<?php
#
# Copyright (c) 2007, 2008 Oligem.com.  All rights reserved.
#
# This code is derived from software contributed to Oligem by
# Marc Vertes <mvertes@free.fr>.
#
# Permission to use, copy, modify, and distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
# OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
#

require('wdim.conf');
require('rsa.php');

$g_script = getenv("SCRIPT_NAME");
$zero_node = "da39a3ee5e6b4b0d3255bfef95601890afd80709";

# create a pending account, email user a request to confirm.
function newuser($name, $mail, $pubkey)
{
	global $g_script, $admin_email;
	$host = $_SERVER['HTTP_HOST'];

	# Missing directories created lazily on the fly.
	if (! is_dir("user")) {
		mkdir("archive");
		mkdir("item");
		mkdir("pending");
		mkdir("user");
	}

	# abort request for already known email
	if (is_file("user/$mail")) {
		print "Error: user present: $mail\n";
		return;
	}
	if (is_file("pending/$mail")) {
		print "Error: user pending: $mail\n";
		return;
	}

	# Create a unique transaction number
	$umd = microtime() . " " . rand() . "\n";
	$umd = md5($umd);

	$msg = "Hello,

This message has been generated and sent by $host$g_script.

User $mail has required an account on
http://$host$g_script. Forward this message to that mail address
to notify your acceptance.

The key for this transaction is $umd

To complete the account creation, $mail must run the shell command:

  \$ dim mkuser $umd

Thanks.";

	mail($admin_email, "dim new user request", $msg,
	     "From: $admin_email\r\n");
	print "The request has been mailed to $admin_email\n";

	$fd = fopen("pending/$umd", "w");
	fputs($fd, "email='$mail'\n");
	fputs($fd, "pubkey='\n$pubkey\n'\n");
	fclose($fd);
	touch("pending/$mail");
}

function fix_buf($buf)
{
	$buf = preg_replace("/^([^=]+)='/", "\$\\1='", $buf);
	$buf = preg_replace("/\n([^=\n]+)='/", "\n$\\1='", $buf);
	$buf = preg_replace("/\'\n/", "';\n", $buf);
	$buf = preg_replace("/\'$/", "';", $buf);
	return $buf;
}

function verify($pubkey, $sign, $target)
{
	# Parse public key (assume 512 bits key)
	$str = preg_replace('/\n*-----.* PUBLIC KEY-----\n*/', "", $pubkey);
	$str = base64_decode($str);
	$str = substr($str, 25, 64);
	$mod = binary_to_number($str);

	# Parse encrypted signature
	$dsign = base64_decode($sign);

	return (rsa_verify($dsign, "65537", $mod, 512) == $target);
}

function newuser2($key, $sname, $sign)
{
	if (! is_readable("pending/$key")) {
		print "Fatal: authentification failed\n";
		return;
	}
	$buf = file_get_contents("pending/$key");
	eval(fix_buf($buf));
	if (! verify($pubkey, $sign, $email)) {
		print "Error: authentication failed\n";
		return;
	}
	if (is_file("user/$sname")) {
		print "Error: invalid shortname: $sname\n";
		return;
	}
	$fd = fopen("pending/$key", 'a');
	fputs($fd, "sname='". $sname . "'\n");
	fclose($fd);
	$fd = fopen("user/$sname", "w");
	fputs($fd, "$email\n");
	fclose($fd);
	unlink("pending/$email");
	if (rename("pending/$key", "user/$email")) {
		print "ok\n";
		return;
	}
}

# Generate a new version name, from the ancestor name.
function newvers($anc, $name, $branch = 0, $main = 0)
{
	if ($main || preg_match('/-' . $name . '[0-9.]+$/', $anc)) {
		preg_match('/([0-9]+)$/', $anc, $match);
		$v2 = $match[1];
		if ($branch) {
			for ($i = 0; $i < $branch - 1; $i++)
				$v2 .= ".0";
			$v2 .= ".1";
		} else
			$v2++;
		$res = preg_replace('/[0-9]+$/', $v2, $anc);
	} else {
		# new initial branch
		$res = $anc  . "-" . $name;
		for ($i = 0; $i < $branch; $i++)
			$res .= "0.";
		if (! preg_match('/[0-9]$/', $res))
			$res .= "1";
	}
	return $res;
}

# new version request processing
function new_version($mail, $sign, $item, $node, $anc, $vname, $bopt,
		     $Mopt, $nopt)
{
	global $zero_node, $admin_email;

	# Check email
	if (! is_readable("user/$mail")) {
		print "Fatal: authentification failed\n";
		return;
	}
	$buf = file_get_contents("user/$mail");
	eval(fix_buf($buf));

	# Check signature
	if (! verify($pubkey, $sign, $node)) {
		print "Error: authentication failed\n";
		return;
	}

	# Check item, or create a new one
	$remove0 = 0;
	if (! is_dir("item/$item")) {
		if ($nopt) $remove0 = 1;
		mkdir("archive/$item");
		mkdir("item/$item");
		mkdir("item/$item/.version");
		mkdir("item/$item/.sum");
		$fd = fopen("item/$item/.sum/$zero_node", "xb");
		if (! $fd) {
			print "Fatal: create sum $zero_node\n";
			return;
		}
		$fd1 = fopen("item/$item/.version/0", "xb");
		if (! $fd1) {
			print "Fatal: create version 0\n";
			fclose($fd);
			return;
		}
		fputs($fd, "0\n");
		fputs($fd1, "$zero_node\n");
		fclose($fd);
		fclose($fd1);
	}

	# Create a new version, first reserve the checksum
	# or return found version if it already exists
	if (! ($fd = fopen("item/$item/.sum/$node", "xb"))) {
		print file_get_contents("item/$item/.sum/$node");
		return;
	}
	# if user is admin, apply a non empty vname
	if ($vname) {
		if ($Mopt) {
			$main = 1;
			$vers = $vname;
			if (! ($fd1 = fopen("item/$item/.version/$vname",
					    "xb"))) {
				print "Fatal: version $vname already exists\n";
				fclose($fd);
				return;
			}
		} else {
			$sname=$vname;
			$branch = 0;
			$main = $Mopt;
			$vers = newvers($anc, $sname, $branch, $main);
			# if proposed name already exists, force a new branch
			while (! ($fd1 = fopen("item/$item/.version/$vers",
					       "xb"))) {
				$branch++;
				if (($vers = newvers($anc, $sname,
						     $branch, $main)) == "") {
					fclose($fd);
					print "Fatal: newvers failed\n";
					return;
				}
			}
		}
	} else {
		# get a new version, append it to the version file,
		# send it back.
		$branch = $bopt;
		$main = $Mopt;
		if (($vers = newvers($anc, $sname, $branch, $main)) == "") {
			fclose($fd);
			print "Fatal: newvers failed\n";
			return;
		}
		# if proposed name already exists, force a new branch
		while (! ($fd1 = fopen("item/$item/.version/$vers", "xb"))) {
			$branch++;
			if (($vers = newvers($anc, $sname,
					     $branch, $main)) == "") {
				fclose($fd);
				print "Fatal: newvers failed\n";
				return;
			}
		}
	}
	if ($nopt) {
		fclose($fd);
		fclose($fd1);
		unlink("item/$item/.version/$vers");
		unlink("item/$item/.sum/$node");
		if ($remove0) {
			unlink("item/$item/.version/0");
			unlink("item/$item/.sum/$zero_node");
			rmdir("item/$item/.version");
			rmdir("item/$item/.sum");
			rmdir("item/$item");
		}
		print "$vers\n";
		return;
	}
	fputs($fd, "$vers\n");
	fputs($fd1, "$node\n");
	fclose($fd);
	fclose($fd1);
	print "$vers\n";
}

# Print the list of file and size in directory
function list_files($dirname)
{
	if (! ($dir = opendir($dirname))) {
		print "Fatal: invalid dir: $dirname\n";
		return;
	}
	while ($file = readdir($dir)) {
		if (is_dir("$dirname/$file")) {
			if ($file == "." || $file == "..")
				continue;
			print "$file\n";
		} else if (is_file("$dirname/$file")) 
			print "$file:" . filesize("$dirname/$file") . "\n";
	}
	closedir($dir);
}

# Process upload file request
function upload_vers($mail, $sign, $item, $dfile, $apfile, $size)
{
	# Check email
	if (! is_readable("user/$mail")) {
		print "Fatal: authentification failed\n";
		return;
	}
	$buf = file_get_contents("user/$mail");
	eval(fix_buf($buf));
	# Check signature
	if (! verify($pubkey, $sign, $dfile)) {
		print "Error: authentication failed\n";
		return;
	}
	# Get file
	move_uploaded_file($_FILES['userfile']['tmp_name'],
			   "archive/$item/$dfile");
	# Check size, discard file if not ok
	if ($size != filesize("archive/$item/$dfile")) {
		unlink("archive/$item/$dfile");
		print "Fatal: invalid size: $dfile\n";
	}
	# append a file chunk to a large file 
	if ($apfile != ".") {
		$fd = fopen("archive/$item/$apfile", "ab");
		fputs($fd, file_get_contents("archive/$item/$dfile"));
		fclose($fd);
		unlink("archive/$item/$dfile");
	}
}

function unexport_version($mail, $sign, $item, $vsum, $nopt)
{
	# Check email
	if (! is_readable("user/$mail")) {
		print "Fatal: authentification failed\n";
		return;
	}
	$buf = file_get_contents("user/$mail");
	eval(fix_buf($buf));
	# Check signature
	if (! verify($pubkey, $sign, $vsum)) {
		print "Error: authentication failed\n";
		return;
	}
	$sumfile = "item/$item/.sum/$vsum";
	if (! file_exists($sumfile)) {
		print "Fatal: invalid version: $vsum\n";
		return;
	}
	$vers = trim(file_get_contents($sumfile));
	print "$vers\n";
	if ($nopt) return;
	unlink("item/$item/.version/$vers");
	unlink($sumfile);
}

# Main
# Global vars from post and forms prefixed by "g_"
error_reporting(E_ERROR | E_PARSE);
import_request_variables("gp", "g_");
header("Content-Type: text/plain");
$nopt = 0;
if (! empty($g_n)) $nopt = $g_n; 

if (! empty($g_nu) && $g_nu == 2) {
	newuser($g_name, $g_mail, $g_pkey);
} else if (! empty($g_nuc)) {
	newuser2($g_key, $g_sn, $g_p);
} else if (! empty($g_nv)) {
	$vn = "";
	$bopt = 0;
	$Mopt = 0;
	$bn = "";
	if (! empty($g_vn) && $g_vn != ".") $vn = $g_vn;
	if (! empty($g_b)) $bopt = 1;
	if (! empty($g_M)) $Mopt = 1;
	if (! empty($g_B)) {
		$bn = $g_B;
		$bopt = 1;
	}
	new_version($g_mail, $g_p, $g_item, $g_nv, $g_anc, $vn,
		    $bopt, $Mopt, $nopt);
} else if (! empty($g_uv)) {
	unexport_version($g_m, $g_p, $g_i, $g_v, $nopt);
} else if (! empty($g_upfile)) {
	upload_vers($g_m, $g_p, $g_item, $g_dfile, $g_af, $g_s);
} else if (! empty($g_ld)) {
	list_files($g_ld);
} 
?>
